/*  wmrt v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>
 *  
 *  wmrt.c -- WindowMaker remote traffic dockapp
 *
 *  This file is part of wmrt.
 *  For more information visit: http://www.eggdrop.ch/projects/wmrt/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* exit codes:
 * 0 - application closed
 * 1 - bad usage
 * 2 - unable to open display
 * 3 - no windowmanager running
 * 4 - can't resolve hostname
 * 5 - socket() failed
 * 6 - can't connect
 * 7 - disconnected
 */

/* 
 * includes
 */

/* ANSI C */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

/* sockets */
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* other */
#include <sys/time.h>

/* X11 */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>


/* 
 * macros 
 */

/* width and height of the dockapp */
#define WMRT_SIZE		64

/* border around the dockapp */
#define WMRT_BORDER		4

/* distance to the graph from top */
#define WMRT_GRAPHDIST		8

/* width of the graph */
#define WMRT_GRAPHWIDTH		(WMRT_SIZE-(2*WMRT_BORDER))

/* height of the graph */
#define WMRT_GRAPHHEIGHT	(WMRT_SIZE-(2*WMRT_BORDER)-WMRT_GRAPHDIST-1)

/* width of the graph without border */
#define WMRT_DATAWIDTH		(WMRT_GRAPHWIDTH-2)

/* height of the graph without border */
#define WMRT_DATAHEIGHT		(WMRT_GRAPHHEIGHT-2)

/* the point on bottom left of the graph */
#define WMRT_BOTLEFT_X		(WMRT_BORDER+1)
#define WMRT_BOTLEFT_Y		(WMRT_BORDER+WMRT_GRAPHDIST+1)

/* the point on bottom right of the graph */
#define WMRT_BOTRIGHT_X		(WMRT_BORDER+WMRT_GRAPHWIDTH-1)
#define WMRT_BOTRIGHT_Y		(WMRT_BORDER+WMRT_GRAPHDIST+WMRT_GRAPHHEIGHT-1)

/* name of the window */
#define WMRT_NAME		"wmrt"

/* name of the applcation */
#define WMRT_CLASS		"WMRT"

/* version of the application */
#define WMRT_VERSION		"1.0"

/* usage */
#define WMRT_USAGE \
	"wmrt version " WMRT_VERSION " - a windowmaker remote traffic " \
	"dockapp\n" \
	"Copyright (C) 2003 by tomasliq <tom@eggdrop.ch>\n" \
	"\n" \
	"Usage: wmrt HOST PORT OPTION BYTES\n" \
	"\n" \
	"HOST is the IP address or the hostname of the rtserver\n" \
	"PORT is the port of the rtserver\n" \
	"OPTION can be -r for received or -t for sent traffic\n" \
	"BYTES is the highest number of bytes per second\n" \
	"\n" \
	"Example for receiving traffic from 192.168.0.1 port 384 with a 64K " \
	"ISDN line:\n" \
	"       wmrt 192.168.0.1 384 -r 8192\n"


/*
 * structures
 */

struct wmrt_draw
{
	/* some xlib variables */
	Display *dpy;
	int screen;
	Colormap cmap;
	Window *win, mwin, iwin;
	GC gc;
	Atom atom;
	int state;

	/* max bytes per second */
	unsigned long maxbytes;

	/* the data of the graph (needed for redrawing)
	 * unsigned char is enough if WMRT_DATAHEIGHT < 256 */
	unsigned char data[WMRT_DATAWIDTH]; 

	/* current traffic in bytes; unsigned long is enough
	 * if we don't have more than 4 GByte/sec traffic */
	unsigned long curtraf;

	int update; /* 1, if received new traffic */

	unsigned char mode; /* 0 = receiving; 1 = sending */

	/* socket stuff */
	struct hostent *host;
	struct sockaddr_in in;
	int sock;
};


void wmrt_setup(struct wmrt_draw *x, char *host, unsigned short port, \
		unsigned char mode, unsigned long maxbytes)
{
	XGCValues gcv;
	XClassHint chint;
	XSetWindowAttributes wattr;
	XSizeHints shints;
	XWMHints hints;
	int i;
	XRectangle bound = { 0, 0, WMRT_SIZE-(2*WMRT_BORDER), \
		WMRT_SIZE-(2*WMRT_BORDER) };

	/* initialize the structure */
	x->state = -1;
	for (i=0;i<WMRT_DATAWIDTH;i++)
		x->data[i] = 0;
	x->curtraf = 0;
	x->update = 0;
	x->mode = mode;
	x->maxbytes = maxbytes;

	/* resolve hostname of the rtserver */
	x->host = gethostbyname(host);
	if (!x->host)
	{
		fprintf(stderr, "Unable to resolve %s\n", host);
		exit(4);
	}

	/* set up socket structure */
	x->in.sin_addr = *((struct in_addr*)x->host->h_addr);
	x->in.sin_family = AF_INET;
	x->in.sin_port = htons(port);
	
	/* create socket */
	x->sock = socket(AF_INET, SOCK_STREAM, 0);
	if (x->sock < 0)
	{
		fprintf(stderr, "socket() failed\n");
		exit(5);
	}

	/* connect to the rtserver */
	if (connect(x->sock, (struct sockaddr*)&x->in, sizeof(x->in)) < 0)
	{
		fprintf(stderr, "Can't connect\n");
		exit(6);
	}

	/* connect to X display */
	if ((x->dpy = XOpenDisplay(NULL)) == NULL)
	{
		fprintf(stderr, "Unable to open display %s\n", \
				XDisplayString(NULL));
		exit(2);
	}

	x->screen = DefaultScreen(x->dpy);

	/* is there a windowmanager running (good would be WindowMaker) */
	x->atom = XInternAtom(x->dpy, "WM_DELETE_WINDOW", False);
	if (x->atom == None)
	{
		fprintf(stderr, "There is no windowmanager running.\n");
		exit(3);
	}

	if (XInternAtom(x->dpy,"_WINDOWMAKER_WM_FUNCTION", True) != None)
	{
		if (x->state == -1) x->state = WithdrawnState;
	}
	else
	{
		if (x->state == -1) x->state = NormalState;
	}


	/* whole stuff for setting up window */

	chint.res_name = WMRT_NAME;
	chint.res_class = WMRT_CLASS;

	wattr.background_pixmap = ParentRelative;

	x->mwin = XCreateSimpleWindow(x->dpy, DefaultRootWindow(x->dpy), \
			0, 0, WMRT_SIZE, WMRT_SIZE, 0, 0, 0);

	XChangeWindowAttributes(x->dpy, x->mwin, CWBackPixmap, &wattr);
	XSetClassHint(x->dpy, x->mwin, &chint);

	gcv.graphics_exposures = 0;
	gcv.foreground = BlackPixel(x->dpy, x->screen);
	gcv.background = WhitePixel(x->dpy, x->screen);
	gcv.font = XLoadFont(x->dpy, "5x8");

	x->gc = XCreateGC(x->dpy, DefaultRootWindow(x->dpy), \
		(GCFont|GCGraphicsExposures|GCForeground|GCBackground), &gcv);

	hints.window_group = x->mwin;
	hints.initial_state = x->state;

	if (x->state == WithdrawnState)
	{
		/* create icon window */
		x->iwin = XCreateSimpleWindow(x->dpy, \
			DefaultRootWindow(x->dpy), \
			0, 0, WMRT_SIZE, WMRT_SIZE, 0, 0, 0);
		XChangeWindowAttributes(x->dpy, x->iwin, CWBackPixmap, &wattr);
		XSetClassHint(x->dpy, x->mwin, &chint);
		x->win = &x->iwin;
		hints.icon_window = x->iwin;
	}
	else
	{
		/* use main window */
		x->win = &x->mwin;
		hints.icon_window = None;
	}

	hints.flags = WindowGroupHint | StateHint | IconWindowHint;
	XSetWMHints(x->dpy,x->mwin,&hints);
	XSetWMProtocols(x->dpy, x->mwin, &x->atom, 1);

	shints.min_width = shints.min_height = shints.max_width = \
		shints.max_height = WMRT_SIZE;

	shints.flags = PMinSize | PMaxSize;
	XSetWMNormalHints(x->dpy, x->mwin, &shints);

	XStoreName(x->dpy, x->mwin, WMRT_NAME);
	XShapeCombineRectangles(x->dpy, *(x->win), ShapeBounding, \
			4, 4, &bound, 1, ShapeBounding, 0);
	XSelectInput(x->dpy, *(x->win), (ExposureMask|ButtonPressMask));
	XMapSubwindows(x->dpy, *(x->win));
	XMapWindow(x->dpy, x->mwin);
}

unsigned long wmrt_getcolor(unsigned char r, unsigned char g, unsigned char b)
{
	return (b+(g*256)+(r*65536));
}

void wmrt_draw(struct wmrt_draw *x)
{
	char str[32];
	int i;

	/* clear window */
	XClearWindow(x->dpy, *(x->win));

	/* draw the graph! */

	/* white border */
	XSetForeground(x->dpy, x->gc, wmrt_getcolor(255,255,255));
	XDrawRectangle(x->dpy, *(x->win), x->gc, \
		WMRT_BORDER, WMRT_BORDER+WMRT_GRAPHDIST, \
		WMRT_GRAPHWIDTH-1, WMRT_GRAPHHEIGHT);

	/* text with traffic */
	if (x->curtraf < 1024)
	{
		snprintf(str, 32, "%ld B/s", x->curtraf);
	}
	else if (x->curtraf < 1048576)
	{
		snprintf(str, 32, "%1.2f KB/s", x->curtraf / 1024.0);
	}
	else if (x->curtraf < 1073741824)
	{
		snprintf(str, 32, "%1.2f MB/s", x->curtraf / 1048576.0);
	}
	else /* wow! */
	{
		snprintf(str, 32, "%1.2f GB/s", x->curtraf / 1073741824.0);
	}

	/* draw it */
	XDrawString(x->dpy, *(x->win), x->gc, 4, 10, str, strlen(str));

	if (x->update) /* new traffic */
	{
		unsigned long t;

		/* magic math with nice cast operators ;-) */
		t = (unsigned long)((double)x->curtraf/((double)x->maxbytes/ \
					(double)(WMRT_DATAHEIGHT+1)));
		
		if (t > WMRT_DATAHEIGHT+1)
			t = WMRT_DATAHEIGHT+1;

		/* update "database" */
		for (i=0;i<(WMRT_DATAWIDTH-1);i++)
			x->data[i]=x->data[i+1];
		x->data[WMRT_DATAWIDTH-1]=t;
		
		x->update=0;
	}

	/* set nice color */
	if (x->mode == 0)
		XSetForeground(x->dpy, x->gc, wmrt_getcolor(0,191,63));
	else
		XSetForeground(x->dpy, x->gc, wmrt_getcolor(0,127,255));

	/* redraw the graph */
	for (i=0;i<WMRT_DATAWIDTH;i++)
	{
		if (x->data[i] == 0)
			continue;
		XDrawLine(x->dpy, *(x->win), x->gc, \
			WMRT_BOTLEFT_X+i, WMRT_BOTRIGHT_Y-x->data[i]+1, \
			WMRT_BOTLEFT_X+i, WMRT_BOTRIGHT_Y);
	}
	
	/* flush the display */
	XFlush(x->dpy);
}

void wmrt_mainloop(struct wmrt_draw *x)
{
	XEvent event;
	int done=0;
	fd_set fds;
	struct timeval tv;
	int b;
	char buffer[32];
	
	tv.tv_sec = 0;
	tv.tv_usec = 0;

	while (!done)
	{
		while (XPending(x->dpy))
		{
			XNextEvent(x->dpy, &event);
			switch (event.type)
			{
				case Expose:
					/* we need to redraw the dockapp */
					wmrt_draw(x);
					break;
				case ClientMessage:
					if(event.xclient.data.l[0] == x->atom)
						done = 1;
					break;
			}
		}

		/* do we have data received? */
		FD_ZERO(&fds);
		FD_SET(x->sock, &fds);
		FD_SET(STDIN_FILENO, &fds);

		b = select(x->sock+1, &fds, NULL, NULL, &tv);
		if (b)
		{
			unsigned long rx, tx;

			/* we received data */
			if ((b = read(x->sock, buffer, sizeof(buffer))) < 1)
			{
				close(x->sock);
				exit(7);
			}
			buffer[b]='\0';
			
			/* parse the data */
			sscanf(buffer, "%lu %lu\r\n", &rx, &tx);

			if (x->mode == 0)
				x->curtraf = rx; /* received */
			else
				x->curtraf = tx; /* transmitted */
			
			x->update = 1;
			
			/* draw it */
			wmrt_draw(x);
		}

		/* this line prevents from eating up all the CPU power */
		usleep(1);
	}
}

int main(int argc, char *argv[])
{
	struct wmrt_draw drw;
	int tmpport;
	unsigned short port;
	unsigned long maxbytes;
	unsigned char mode;

	if (argc != 5)
	{
		fprintf(stderr, "%s", WMRT_USAGE);
		exit(1);
	}

	/* check port */
	tmpport = atoi(argv[2]);
	if (tmpport < 1 || tmpport > 65535)
	{
		fprintf(stderr, "%s", WMRT_USAGE);
		exit(1);
	}
	port = (unsigned short)tmpport;


	/* receive or send? */
	/* TODO: both receiving and sending; less strict usage */
	mode = 2;
	if (!strcmp(argv[3], "-r")) mode = 0;
	if (!strcmp(argv[3], "-t")) mode = 1;
	if (mode == 2)
	{
		fprintf(stderr, "%s", WMRT_USAGE);
		exit(1);
	}
	maxbytes = strtoul(argv[4], (char **)NULL, 10); /* ANSI C 99 */
	/* if your compiler doesn't support ANSI C 99 change the above line to:
	 * maxbytes = atoi(argv[4]);
	 */

	/* TODO: display "Connecting..." on the dockapp if we aren't connected
	 *       yet. show an error message on the dockapp if connecting
	 *       failed. also try to reconnect if we disconnected or can't
	 *       connect instead of exiting.
	 */
	
	wmrt_setup(&drw, argv[1], port, mode, maxbytes);
	wmrt_mainloop(&drw);
	exit(0);
}
